"""
Custom exceptions used in this library.
"""

class SetIntersectionError(Exception):
    """Happens when the intersection of two sets is non-empty."""

    pass


class NonEmptyError(Exception):
    """Happens when something that is supposed to be empty is not."""

    pass


class NotACharacterException(Exception):
    """Happens when something that should have been a character is not a character."""

    pass



"""
Convenient constants for use in Regex.
"""

#: Any whitespace character
WHITESPACE = r"\s"

#: Any non-whitespace character
NON_WHITESPACE = r"\S"

#: Any lowercase character a-z
LOWERCASE = r"[a-z]"

#: Any uppercase character A-Z
UPPERCASE = r"[A-Z]"

#: Any lower or uppercase character a-z A-Z
ALPHA = r"[a-zA-Z]"

#: Any numeric character 0-9
NUMERIC = r"\d"

#: Any non-numeric character
NON_NUMERIC = r"\D"

#: Any alphanumeric character a-z or 0-9
ALPHA_NUMERIC = r"\w"

#: Any non-alphanumeric character
NON_ALPHANUMERIC = r"\W"

#: Any character at all
ANY = r"."

#: The literal period
PERIOD = r"\."

#: The start of a line
START_OF_LINE = r"^"

#: The end of a line
END_OF_LINE = r"$"

#: One or more of the last character or group
ONE_OR_MORE = r"+"

#: Zero or more of the last character or group
ZERO_OR_MORE = r"*"

#: Zero or one of the last character or group
OPTIONAL = r"?"

#: Literal Plus
PLUS = r"\+"

#: Literal ASTERISK
ASTERISK = r"\*"

#: Literal Dollar Sign
DOLLAR = r"\$"

#: Literal Question Mark
QUESTION_MARK = r"\?"

#: Indicates a newline, use in `re.MULTILINE` mode.
#: Supports all operating systems
NEWLINE = r"(?:\n|\r\n?)"



"""
Tests Regex class by testing output string representation.
"""

from typing import Callable

import pytest

from iregex import Regex
from iregex.consts import (
    ALPHA,
    ANY,
    NUMERIC,
    ONE_OR_MORE,
    NEWLINE,
    WHITESPACE,
    ZERO_OR_MORE,
)
from iregex.exceptions import NonEmptyError, SetIntersectionError, NotACharacterException


def test_regex_literal() -> None:
    """Test basic literal addition."""
    regex = Regex(NUMERIC).literal(ALPHA)
    assert str(regex) == NUMERIC + ALPHA


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).whitespace(), NUMERIC + WHITESPACE + ZERO_OR_MORE),
        (Regex(NUMERIC+ALPHA).whitespace(), NUMERIC + ALPHA + WHITESPACE + ZERO_OR_MORE),
    ],
)
def test_whitespace(regex: Regex, result: str) -> None:
    """Test basic whitespace addition."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).newline(), NUMERIC + NEWLINE + ZERO_OR_MORE),
        (Regex(NUMERIC+ALPHA).newline(), NUMERIC + ALPHA + NEWLINE + ZERO_OR_MORE),
    ],
)
def test_newline(regex: Regex, result: str) -> None:
    """Test basic newline addition."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).zero_or_more_repetitions(), f"{NUMERIC}" + ZERO_OR_MORE),
        (Regex(NUMERIC+ALPHA).zero_or_more_repetitions(), f"(?:{NUMERIC+ALPHA})" + ZERO_OR_MORE),
    ],
)
def test_zero_or_more_repetitions(regex: Regex, result: str) -> None:
    """Test basic repetitions."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).one_or_more_repetitions(), f"{NUMERIC}" + ONE_OR_MORE),
        (Regex(NUMERIC+ALPHA).one_or_more_repetitions(), f"(?:{NUMERIC+ALPHA})" + ONE_OR_MORE),
    ],
)
def test_one_or_more_repetitions(regex: Regex, result: str) -> None:
    """Test basic repetitions."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).m_to_n_repetitions(3, 5), f"{NUMERIC}" + "{3,5}"),
        (Regex(NUMERIC+ALPHA).m_to_n_repetitions(3, 5), f"(?:{NUMERIC+ALPHA})" + "{3,5}"),
    ],
)
def test_m_to_n_repetitions(regex: Regex, result: str) -> None:
    """Test basic repetitions."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).exactly_m_repetitions(3), f"{NUMERIC}" + "{3}"),
        (Regex(NUMERIC+ALPHA).exactly_m_repetitions(3), f"(?:{NUMERIC+ALPHA})" + "{3}"),
    ],
)
def test_exactly_m_repetitions(regex: Regex, result: str) -> None:
    """Test basic repetitions."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).m_or_more_repetitions(3), f"{NUMERIC}" + "{2}" + f"{NUMERIC}" + ONE_OR_MORE),
        (Regex(NUMERIC+ALPHA).m_or_more_repetitions(3), f"(?:{NUMERIC+ALPHA})" + "{2}" + f"(?:{NUMERIC+ALPHA})" + ONE_OR_MORE),
    ],
)
def test_m_or_more_repetitions(regex: Regex, result: str) -> None:
    """Test basic repetitions."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex(NUMERIC).optional(), f"{NUMERIC}?"),
        (Regex(NUMERIC+ALPHA).optional(), f"(?:{NUMERIC+ALPHA})?"),
    ],
)
def test_optional(regex: Regex, result: str) -> None:
    """Test basic optional."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex().any_char("a"), f"a"),
        (Regex().any_char("a", "b"), f"[ab]"),
    ],
)
def test_any_char(regex: Regex, result: str) -> None:
    """Test basic any_char."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex_lazy",
    [
        lambda: Regex().any_char("ab"),
        lambda: Regex().any_char("a", "bc"),
    ],
)
def test_any_char_error(regex_lazy: Callable) -> None:
    """Test basic any_char."""
    with pytest.raises(NotACharacterException):
        regex_lazy()


@pytest.mark.parametrize(
    "regex,result",
    [
        (Regex().exclude_char("a"), f"[^a]"),
        (Regex().exclude_char("a", "b"), f"[^ab]"),
    ],
)
def test_exclude_char(regex: Regex, result: str) -> None:
    """Test basic exclude_char."""
    assert str(regex) == result


@pytest.mark.parametrize(
    "regex_lazy",
    [
        lambda: Regex().exclude_char("ab"),
        lambda: Regex().exclude_char("a", "bc"),
    ],
)
def test_exclude_char_error(regex_lazy: Callable) -> None:
    """Test basic exclude_char."""
    with pytest.raises(NotACharacterException):
        regex_lazy()


def test_anything() -> None:
    """Test basic anything."""
    regex = Regex(NUMERIC).anything()
    assert str(regex) == NUMERIC + ANY + ZERO_OR_MORE


def test_capture_group1() -> None:
    """Basic test capture group."""
    regex = Regex(NUMERIC + ALPHA).make_capture_group().literal(ZERO_OR_MORE)
    assert str(regex) == f"({NUMERIC+ALPHA}){ZERO_OR_MORE}"


def test_non_capture_group1() -> None:
    """Basic test capture group."""
    regex = Regex(NUMERIC + ALPHA).make_non_capture_group().literal(ZERO_OR_MORE)
    assert str(regex) == f"(?:{NUMERIC+ALPHA}){ZERO_OR_MORE}"


def test_named_capture_group1() -> None:
    """Basic test named capture group."""
    regex = (
        Regex(NUMERIC + ALPHA).make_named_capture_group("name").literal(ZERO_OR_MORE)
    )
    assert str(regex) == f"(?<name>{NUMERIC+ALPHA}){ZERO_OR_MORE}"


def test_make_lookahead() -> None:
    """Test basic lookahead."""
    regex = Regex(ALPHA).literal(Regex(NUMERIC).make_lookahead())
    assert str(regex) == ALPHA + f"(?={NUMERIC})"


def test_make_negative_lookahead() -> None:
    """Test negative lookahead."""
    regex = Regex(ALPHA).literal(Regex(NUMERIC).make_negative_lookahead())
    assert str(regex) == ALPHA + f"(?!{NUMERIC})"


def test_make_lookbehind() -> None:
    """Test basic lookbehind."""
    regex = Regex(NUMERIC).make_lookbehind().literal(ALPHA)
    assert str(regex) == f"(?<={NUMERIC})" + ALPHA


def test_make_negative_lookbehind() -> None:
    """Test negative lookbehind."""
    regex = Regex(NUMERIC).make_negative_lookbehind().literal(ALPHA)
    assert str(regex) == f"(?<!{NUMERIC})" + ALPHA


@pytest.mark.parametrize(
    "self,other,result",
    [
        (Regex(NUMERIC), Regex(NUMERIC), Regex(NUMERIC).literal(NUMERIC)),
        (Regex(NUMERIC), NUMERIC, Regex(NUMERIC).literal(NUMERIC)),
        (NUMERIC, Regex(NUMERIC), Regex(NUMERIC).literal(NUMERIC)),
        (
            Regex(NUMERIC).make_named_capture_group("name1"),
            Regex(ALPHA).make_named_capture_group("name2"),
            Regex(f"(?<name1>{NUMERIC})(?<name2>{ALPHA})"),
        ),
    ],
)
def test_add(self: Regex, other: Regex, result: Regex) -> None:
    """Tests lots of adds."""
    assert self + other == result


@pytest.mark.parametrize(
    "self,other",
    [
        (
            Regex(NUMERIC).make_named_capture_group("name"),
            Regex(ALPHA).make_named_capture_group("name"),
        )
    ],
)
def test_add_set_intersection_error(self: Regex, other: Regex) -> None:
    """Tests that an add error pops up in certain scenarios."""
    with pytest.raises(SetIntersectionError):
        self + other


@pytest.mark.parametrize(
    "self,other,result",
    [
        (Regex(NUMERIC), Regex(NUMERIC), f"(?:{NUMERIC}|{NUMERIC})"),
        (Regex(NUMERIC), NUMERIC, f"(?:{NUMERIC}|{NUMERIC})"),
        (NUMERIC, Regex(NUMERIC), f"(?:{NUMERIC}|{NUMERIC})"),
    ],
)
def test_or(self: Regex, other: Regex, result: Regex) -> None:
    """Tests basic or."""
    assert str(self | other) == str(result)


@pytest.mark.parametrize(
    "self,other",
    [
        (
            Regex(NUMERIC).make_named_capture_group("name1"),
            Regex(ALPHA).make_named_capture_group("name2"),
        )
    ],
)
def test_or_non_empty_error(self: Regex, other: Regex) -> None:
    """Tests that an or error pops up in certain scenarios."""
    with pytest.raises(NonEmptyError):
        self | other



"""
This is the module containing the main Regex class.
"""

import re
from copy import copy
from typing import List, Optional, Union, Any

from iregex.consts import ANY, ONE_OR_MORE, OPTIONAL, WHITESPACE, ZERO_OR_MORE, NEWLINE
from iregex.exceptions import NonEmptyError, SetIntersectionError, NotACharacterException


class Regex:
    """
    A wrapper for regex strings that hides the implementation.

    This class is immutable by convention.
    """

    # Private Variables
    _data: List[str]
    _capture_groups: List[str]

    def __init__(self, regex_str: Optional[str] = None) -> None:
        """
        Optionally can take a literal as input.

        :param regex_str: An optional literal to start your Regex with.
        """
        self._data = [regex_str] if regex_str else []
        self._capture_groups = []

    # ============== Chained Methods ==============
    def literal(self, regex: Union[str, "Regex"]) -> "Regex":
        """
        Adds a literal to the end of the regex.

        :param regex: The regex to append to this regex.
        :raises TypeError: When regex is not type Regex or str.
        """
        out = copy(self)
        if isinstance(regex, Regex):
            out._data += regex._data
        elif isinstance(regex, str):
            out._data.append(regex)
        else:
            raise TypeError(f"Type not supported: {type(regex)}")
        return out

    def anything(self) -> "Regex":
        """Appends zero or more of any character to the Regex."""
        return self.literal(ANY + ZERO_OR_MORE)

    def whitespace(self) -> "Regex":
        """Allows unlimited whitespace."""
        out = copy(self)
        out._data.append(WHITESPACE + ZERO_OR_MORE)
        return out

    def newline(self) -> "Regex":
        """Allows newline."""
        out = copy(self)
        out._data.append(NEWLINE + ZERO_OR_MORE)
        return out

    def zero_or_more_repetitions(self) -> "Regex":
        """Repeats the previous regex zero or more times."""
        out = self.make_non_capture_group()
        out._data.append(ZERO_OR_MORE)
        return out

    def one_or_more_repetitions(self) -> "Regex":
        """Repeats the previous regex one or more times."""
        out = self.make_non_capture_group()
        out._data.append(ONE_OR_MORE)
        return out

    def m_to_n_repetitions(self, m: int, n: int) -> "Regex":
        """
        Repeats the previous regex m to n inclusive times.

        :param m: At least this many times.
        :param n: At most this many times (inclusive).
        """
        out = self.make_non_capture_group()
        out._data.append("{" + str(m) + "," + str(n) + "}")
        return out

    def exactly_m_repetitions(self, m: int) -> "Regex":
        """
        Repeats the previous regex exactly m times.

        :param m: Exactly this many instances.
        """
        out = self.make_non_capture_group()
        out._data.append("{" + str(m) + "}")
        return out

    def m_or_more_repetitions(self, m: int) -> "Regex":
        """
        Repeats the previous regex m or more times.

        :param m: This many or more instances.
        :raises ValueError: When m < 0
        """
        if m == 0:
            return self.zero_or_more_repetitions()
        elif m == 1:
            return self.one_or_more_repetitions()
        elif m < 0:
            raise ValueError(f"m must be >= 0, got {m}")
        return self.exactly_m_repetitions(m - 1) + self.one_or_more_repetitions()

    def optional(self) -> "Regex":
        """The previous regex can exist 0 or 1 times."""
        out = self.make_non_capture_group()
        out._data.append(OPTIONAL)
        return out

    def any_char(self, *char: str) -> "Regex":
        """
        Any char in the text may be used.

        :param char: Some character.
        :raises NotACharacterException: Raised if any argument is not a character.
        """
        out = copy(self)
        for t in char:
            if not Regex._is_character(t):
                raise NotACharacterException(f"{t} is not a character.")
        if len(char) > 1:
            out._data += ["["] + list(char) + ["]"]
        else:
            out._data = list(char)
        return out

    def exclude_char(self, *char: str) -> "Regex":
        """
        None of the chars in the text may be used.

        :param char: Some character.
        :raises NotACharacterException: Raised if any argument is not a character.
        """
        out = copy(self)
        for t in char:
            if not Regex._is_character(t):
                raise NotACharacterException(f"{t} is not a character.")
        out._data += ["[^"] + list(char) + ["]"]
        return out

    def make_capture_group(self) -> "Regex":
        """Creates an anonymous capture group."""
        out = copy(self)
        out._data = ["("] + out._data + [")"]
        return out

    @staticmethod
    def _is_character(txt: Union[str, "Regex"]) -> bool:
        """Tests if a regex string is just a single character."""
        txt_: str = str(txt) if isinstance(txt, Regex) else txt
        if len(txt_) == 1:
            return True
        if len(txt_) == 2 and txt_[0] == '\\':
            return True
        return False

    @staticmethod
    def _is_character_group(txt: Union[str, "Regex"]) -> bool:
        """Tests if a regex string is a character group."""
        txt_: str = str(txt) if isinstance(txt, Regex) else txt
        if len(txt_) >= 2 and txt_[0] == r'[' and txt_[-1] == r']':
            return True
        return False

    @staticmethod
    def _is_capture_group(txt: Union[str, "Regex"]) -> bool:
        """Tests if a regex string is a capture group."""
        txt_: str = str(txt) if isinstance(txt, Regex) else txt
        if len(txt_) >= 2 and txt_[0] == r'(' and txt_[-1] == r')':
            return True
        return False

    def make_non_capture_group(self) -> "Regex":
        """Create a capture group that you don't care to retreive the contents of."""
        out = copy(self)
        # You don't need to make a non_capture_group for simple cases
        if Regex._is_character(self) or \
                Regex._is_character_group(self) or \
                Regex._is_capture_group(self):
            out._data = out._data
        else:
            out._data = ["(?:"] + out._data + [")"]
        return out

    def make_named_capture_group(self, name: str) -> "Regex":
        """
        Makes the previous regex a capture group of name `name`.

        :param name: The name to assign the capture group.
        """
        out = copy(self)
        out._data = ["(?<" + name + ">"] + out._data + [")"]
        out._capture_groups.append(name)
        return out

    def make_lookahead(self) -> "Regex":
        """Makes the previous regex a lookahead group."""
        out = copy(self)
        out._data = ["(?="] + self._data + [")"]
        return out

    def make_lookbehind(self) -> "Regex":
        """Makes the previous regex a lookbehind group."""
        out = copy(self)
        out._data = ["(?<="] + self._data + [")"]
        return out

    def make_negative_lookahead(self) -> "Regex":
        """Makes the previous regex a negative lookahead group."""
        out = copy(self)
        out._data = ["(?!"] + self._data + [")"]
        return out

    def make_negative_lookbehind(self) -> "Regex":
        """Makes the previous regex a negative lookbehind group."""
        out = copy(self)
        out._data = ["(?<!"] + self._data + [")"]
        return out

    # ============== Result Methods =============
    def compile(self, *args: Any, **kwargs: Any) -> re.Pattern:
        """
        A simple wrapper around re.compile.

        :param args: Positional parameters to pass to re.compile.
        :param kwargs: Keyword parameters to pass to re.compile.
        """
        return re.compile(str(self), *args, **kwargs)

    # ============== Magic Methods ==============
    def __copy__(self) -> "Regex":
        """A simple copy command."""
        out = Regex()
        out._data = copy(self._data)
        out._capture_groups = copy(self._capture_groups)
        return out

    def __str__(self) -> str:
        """Converts this object into a regex string."""
        return "".join(self._data)

    def __repr__(self) -> str:
        """For debugging and printing."""
        return "Regex(" + str(self) + ")"

    def __eq__(self, other: "Regex") -> bool:  # type: ignore
        """Two Regex's are equal if their regex strings are equal."""
        return str(self) == str(other)

    def __add__(self, other: Union["Regex", str]) -> "Regex":
        """
        Adding two Regex's is just appending their strings.
        But they aren't allowed to share capture group names.
        Can be used as an easy substitute for literal. But is a little slower.

        :raises SetIntersectionError: If the two _capture_groups share any values.
        :raises TypeError: If other is not either a Regex or a string.
        """
        # Handle data types
        if isinstance(other, str):
            other_ = Regex(other)
        elif isinstance(other, Regex):
            other_ = other
            if set(self._capture_groups) & set(other._capture_groups):
                raise SetIntersectionError(
                    "Capture groups in self and other have common names."
                )
        else:
            raise TypeError(f"Unrecognized type: {type(other)}")
        del other  # So you don't reuse the variable

        out = Regex()
        out._data = self._data + other_._data
        out._capture_groups = self._capture_groups + other_._capture_groups
        return out

    def __radd__(self, other: Union["Regex", str]) -> "Regex":
        """
        Adding two Regex's is just appending their strings.
        But they aren't allowed to share capture group names.
        Can be used as an easy substitute for literal. But is a little slower.

        :raises SetIntersectionError: If the two _capture_groups share any values.
        :raises TypeError: If other is not either a Regex or a string.
        """
        # Handle data types
        if isinstance(other, str):
            other_ = Regex(other)
        elif isinstance(other, Regex):
            other_ = other
            if set(self._capture_groups) & set(other._capture_groups):
                raise SetIntersectionError(
                    "Capture groups in self and other have common names."
                )
        else:
            raise TypeError(f"Unrecognized type: {type(other)}")
        del other  # So you don't reuse the variable

        out = Regex()
        out._data = other_._data + self._data
        out._capture_groups = other_._capture_groups + self._capture_groups
        return out

    def __or__(self, other: Union[str, "Regex"]) -> "Regex":
        """
        The `or` of two Regex's is the group `(self|other)`.
        Neither self nor other may contained named capture groups.

        :raises NonEmptyError: If either contains named capture groups.
        """
        # Handle data types
        if isinstance(other, str):
            other_ = Regex(other)
        elif isinstance(other, Regex):
            other_ = other
            if other._capture_groups:
                raise NonEmptyError(
                    f"Capture groups in other is not empty. Found: {other._capture_groups}"
                )
        else:
            raise TypeError(f"Unrecognized type: {type(other)}")
        del other  # So you don't reuse the variable

        # Some other errors
        if self._capture_groups:
            raise NonEmptyError(
                f"Capture groups in self is not empty. Found: {self._capture_groups}"
            )

        out = Regex()
        out._data = ["(?:"] + self._data + ["|"] + other_._data + [")"]
        out._capture_groups = []
        return out

    def __ror__(self, other: Union[str, "Regex"]) -> "Regex":
        """
        The `or` of two Regex's is the group `(self|other)`.
        Neither self nor other may contained named capture groups.

        :raises NonEmptyError: If either contains named capture groups.
        """
        # Handle data types
        if isinstance(other, str):
            other_ = Regex(other)
        elif isinstance(other, Regex):
            other_ = other
            if other._capture_groups:
                raise NonEmptyError(
                    f"Capture groups in other is not empty. Found: {other._capture_groups}"
                )
        else:
            raise TypeError(f"Unrecognized type: {type(other)}")
        del other  # So you don't reuse the variable

        # Some other errors
        if self._capture_groups:
            raise NonEmptyError(
                f"Capture groups in self is not empty. Found: {self._capture_groups}"
            )

        out = Regex()
        out._data = ["(?:"] + other_._data + ["|"] + self._data + [")"]
        out._capture_groups = []
        return out



"""
Tests Regex class by testing the regex match results on given strings.
"""

import pytest

from iregex import Regex
from iregex.consts import ALPHA, NUMERIC


@pytest.mark.parametrize("text,expected", [("1a", True), ("1", False)])
def test_regex_literal_results(text: str, expected: bool) -> None:
    """Test basic literal addition."""
    regex = Regex(NUMERIC).literal(ALPHA).compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected", [("1", True), ("1 ", True), ("1  ", True), (" 1 ", False)]
)
def test_whitespace_results(text: str, expected: bool) -> None:
    """Test basic whitespace addition."""
    regex = Regex(NUMERIC).whitespace().compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected", [("1\n", True),
                      ("1\n\r", True),
                      ("1\r", True)]
)
def test_newline_results(text: str, expected: bool) -> None:
    """Test basic newline addition."""
    regex = Regex(NUMERIC).newline().compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected", [("", True), ("1", True), ("12", True), ("1a", False)]
)
def test_zero_or_more_repetitions_results(text: str, expected: bool) -> None:
    """Test basic repetitions."""
    regex = Regex(NUMERIC).zero_or_more_repetitions().compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected",
    [
        ("", False),
        ("1", True),
        ("12", True),
        ("123", True),
        ("12345", True),
        ("123456", True),
        ("1a", False),
    ],
)
def test_one_or_more_repetitions_results(text: str, expected: bool) -> None:
    """Test basic repetitions."""
    regex = Regex(NUMERIC).one_or_more_repetitions().compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected,m,n",
    [
        ("", False, 3, 5),
        ("1", False, 3, 5),
        ("12", False, 3, 5),
        ("123", True, 3, 5),
        ("12345", True, 3, 5),
        ("123456", False, 3, 5),
        ("1a", False, 3, 5),
    ],
)
def test_m_to_n_repetitions_results(text: str, expected: bool, m: int, n: int) -> None:
    """Test basic repetitions."""
    assert m < n
    regex = Regex(NUMERIC).m_to_n_repetitions(m, n).compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected,m",
    [
        ("", False, 3),
        ("1", False, 3),
        ("12", False, 3),
        ("123", True, 3),
        ("12345", False, 3),
        ("123456", False, 3),
        ("1a", False, 3),
    ],
)
def test_exactly_m_repetitions_results(text: str, expected: bool, m: int) -> None:
    """Test basic repetitions."""
    regex = Regex(NUMERIC).exactly_m_repetitions(m).compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)


@pytest.mark.parametrize(
    "text,expected,m",
    [
        ("", False, 3),
        ("1", False, 3),
        ("12", False, 3),
        ("123", True, 3),
        ("12345", True, 3),
        ("123456", True, 3),
        ("1a", False, 3),
    ],
)
def test_m_or_more_repetitions_results(text: str, expected: bool, m: int) -> None:
    """Test basic repetitions."""
    regex = Regex(NUMERIC).m_or_more_repetitions(m).compile()
    if expected:
        assert regex.fullmatch(text)
    else:
        assert not regex.fullmatch(text)
